# Before `make install' is performed this script should be runnable with
# `make test'. After `make install' it should work as `perl 001.t'

#########################

use Test;

plan tests => 1;

use JavaScript::Squish;
ok(1); # If we made it this far, we're ok.


